(function ($, Drupal) {

  Drupal.ajaxProgressClasses = {
    element: 'ajax-loading',
    form: 'form--ajax-loading',
    modalBody: 'page--dialog-loading'
  };

  /**
   * Add ajax helper classes.
   */
  function addAjaxHelperClasses(element) {
    var $element = $(element);
    $element.addClass(Drupal.ajaxProgressClasses.element);
    $element.closest('form').addClass(Drupal.ajaxProgressClasses.form);
    if ($element.data('dialog-type') == 'modal') {
      $('body').addClass(Drupal.ajaxProgressClasses.modalBody);
    }
  }

  /**
   * Remove ajax helper classes.
   */
  function removeHelperClasses(element) {
    var $element = $(element);
    $element.removeClass(Drupal.ajaxProgressClasses.element);
    $element.closest('form').removeClass(Drupal.ajaxProgressClasses.form);
    if ($element.data('dialog-type') == 'modal') {
      $('body').removeClass(Drupal.ajaxProgressClasses.modalBody);
    }
  }

  /**
   * Override AJAX "beforeSerialize" callback.
   */
  var originalAjaxBeforeSerialize = Drupal.Ajax.prototype.beforeSerialize;
  Drupal.Ajax.prototype.beforeSerialize = function (element, options) {
    var $triggeredElement = $(this.element);

    // Change request url to url from element data-ajax-url attribute
    var elementAjaxUrl = $triggeredElement.data('ajax-url');
    if (elementAjaxUrl) {
      var wrapperFormatIndex = options.url.indexOf(Drupal.ajax.WRAPPER_FORMAT);
      if (wrapperFormatIndex > 0) {
        elementAjaxUrl += (elementAjaxUrl.indexOf('?') === -1) ? '?' : '&';
        elementAjaxUrl += options.url.substring(wrapperFormatIndex);
      }
      options.url = elementAjaxUrl;
    }

    // Trigger custom event
    $triggeredElement.trigger('ajaxBeforeSerialize', [this, options]);

    // Call original callback
    originalAjaxBeforeSerialize.apply(this, arguments);
  };

  /**
   * Override AJAX "beforeSend" callback.
   */
  var originalAjaxBeforeSend = Drupal.Ajax.prototype.beforeSend;
  Drupal.Ajax.prototype.beforeSend = function (xmlhttprequest, options) {
    var $element = $(this.element);
    var elementFocused = $element.is(':focus');

    // Add custom css classes
    addAjaxHelperClasses($element);

    // Set progress type from "data-progress-type" attribute
    // @TODO Remove after close issue https://www.drupal.org/project/drupal/issues/2818463
    if ($element.data('progress-type')) {
      this.progress.type = $element.data('progress-type');
    }

    // Trigger custom event
    $element.trigger('ajaxBeforeSend', [this, options]);

    // Call original callback
    originalAjaxBeforeSend.apply(this, arguments);

    // Remove "disabled" attribute from input
    if ('disabled' in this && !this.disabled) {
      this.ajaxing = false;
      $element.prop('disabled', false);

      // Restore focus
      if (elementFocused) {
        $element.focus();
      }

      // Abort previous ajax request
      if (this.$form && this.$form.data('jqxhr')) {
        this.$form.data('jqxhr').abort();
      }
    }
  };

  /**
   * Override AJAX "success" callback.
   */
  var originalAjaxSuccess = Drupal.Ajax.prototype.success;
  Drupal.Ajax.prototype.success = function (response, status) {
    // Remove custom css classes
    removeHelperClasses(this.element);

    // Call original callback
    originalAjaxSuccess.apply(this, arguments);
  };

  /**
   * Override AJAX "error" callback.
   */
  var originalAjaxError = Drupal.Ajax.prototype.error;
  Drupal.Ajax.prototype.error = function (response, status) {
    // Remove custom css classes
    removeHelperClasses(this.element);

    // Call original callback
    originalAjaxError.apply(this, arguments);
  };

  /**
   * Override AJAX "eventResponse" function.
   */
  var originalAjaxEventResponse = Drupal.Ajax.prototype.eventResponse;
  Drupal.Ajax.prototype.eventResponse = function (element, event) {
    // AJAX "once" functionality
    var $element = $(element);
    if ($element.hasClass('use-ajax-once')) {
      if ($element.data('ajax-done')) {
        this.ajaxing = true;
      }
      else {
        $element.data('ajax-done', true);
      }
    }

    // Call original callback
    originalAjaxEventResponse.apply(this, arguments);
  };

  /**
   * Override "Insert" command.
   */
  var originalInsertCommand = Drupal.AjaxCommands.prototype.insert;
  Drupal.AjaxCommands.prototype.insert = function (ajax, response, status) {
    // Trim data
    if (response.data) {
      response.data = $.trim(response.data);
    }

    // Call original command
    originalInsertCommand.apply(this, arguments);
  };

  /**
   * Command "historyReplaceState".
   */
  Drupal.AjaxCommands.prototype.historyReplaceState = function (ajax, response, status) {
    history.replaceState(null, null, response.url);
  };

  /**
   * Command "pageReload".
   */
  Drupal.AjaxCommands.prototype.pageReload = function (ajax, response, status) {
    location.reload();
  };

  /**
   * Command "consoleLog".
   */
  Drupal.AjaxCommands.prototype.consoleLog = function (ajax, response, status) {
    console.log(response.message);
  };

  /**
   * Command "formReset".
   */
  Drupal.AjaxCommands.prototype.formReset = function (ajax, response, status) {
    var $form = $(response.selector);

    $form.trigger('reset');

    if (response.removeErrors) {
      $form.find('[data-drupal-messages]').remove();
      $form.find(':input.error').removeClass('error');
    }
  };

  /**
   * Create custom progress type "toggleClass".
   * @see Drupal.Ajax.prototype.beforeSend()
   */
  Drupal.Ajax.prototype.setProgressIndicatorToggleclass = function () {
    var $progressTarget = this.progress.target ? $(this.progress.target) : $(this.element);
    var progressClass = this.progress.class ? this.progress.class : 'ajax-progress-animation';

    // Add class
    $progressTarget.addClass(progressClass);

    // Remove class
    this.progress.object = {
      stopMonitoring: function () {
        $progressTarget.removeClass(progressClass);
      }
    };
  };

  /**
   * Send Ajax request and execute response commands using Drupal Ajax API.
   */
  Drupal.ajaxExecute = function (url, options) {
    options = options || {};
    if (url) {
      options.url = url;
    }
    var ajax = new Drupal.Ajax(false, false, options);
    return ajax.execute();
  }

})(jQuery, Drupal);
