<?php

use Drupal\Component\Utility\Html;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\Markup;
use Drupal\Core\Template\Attribute;
use Drupal\views\ViewExecutable;

/**
 * @see improvements_theme()
 */
function improvements_form_theme(): array {
  return [
    // Override system "select" template for add support attributes for "option" element
    'select' => [
      'render element' => 'element',
    ],
  ];
}

/**
 * @see improvements_library_info_alter()
 */
function improvements_form_library_info_alter(array &$libraries, string $extension): void {
  static $improvements_assets_path;

  if (!$improvements_assets_path) {
    $improvements_assets_path = '/' . drupal_get_path('module', 'improvements') . '/assets/';
  }

  // core
  if ($extension == 'core') {
    $libraries['jquery.form']['js'][$improvements_assets_path . 'improvements.form.js'] = [];
    $libraries['jquery.form']['dependencies'][] = 'core/drupal.debounce';
  }
  // file
  elseif ($extension == 'file') {
    $libraries['drupal.file']['js'][$improvements_assets_path . 'improvements.file.js'] = [];
  }
}

/**
 * @see improvements_element_info_alter()
 */
function improvements_form_element_info_alter(array &$elements): void {
  // Improvements managed_file element functionality
  $elements['managed_file']['#pre_render'][] = '\Drupal\improvements\ImprovementsTrustedCallbacks::managedFilePreRender';

  // Improvements textarea element functionality
  $elements['textarea']['#attached']['library'][] = 'improvements/textarea';
}

/**
 * Implements hook_form_alter().
 */
function improvements_form_alter(array &$form, FormStateInterface $form_state, string $form_id): void {
  // Add hidden value to forms in Dialog
  if (in_array(\Drupal::request()->query->get('_wrapper_format'), ['drupal_modal', 'drupal_dialog'])) {
    $form['_in_dialog'] = [
      '#type' => 'hidden',
      '#default_value' => 1,
      '#weight' => 10000,
    ];
  }

  foreach (['form_build_id', 'form_token', 'form_id'] as $element_name) {
    if (isset($form[$element_name])) {
      // Move hidden system elements to form bottom
      $form[$element_name]['#weight'] = 10000;

      // Remove autocomplete attribute from system elements
      unset($form[$element_name]['#attributes']['autocomplete']);
    }
  }

  // Add class to views exposed form
  if ($form_id == 'views_exposed_form') {
    $view_executable = $form_state->get('view'); /** @var ViewExecutable $view_executable */
    $form['#attributes']['class'][] = Html::getClass($view_executable->id()) . '-views-exposed-form';
  }

  improvements_form_tracking_submit_form_alter($form, $form_state, $form_id);
}

/**
 * Preprocess function for captcha.html.twig.
 */
function improvements_preprocess_captcha(array &$vars): void {
  // Attach js for improvements recaptcha functionality
  if (
    isset($vars['element']['#captcha_validate']) &&
    $vars['element']['#captcha_validate'] == 'recaptcha_captcha_validation'
  ) {
    $vars['#attached']['library'][] = 'improvements/recaptcha';
  }
}

/**
 * Preprocess function for input.html.twig.
 */
function improvements_preprocess_input(array &$vars): void {
  // Attach improvements/spinner library if input has class "form-spinner".
  if (!empty($vars['attributes']['class']) && in_array('form-spinner', $vars['attributes']['class'])) {
    $vars['#attached']['library'][] = 'improvements/spinner';
  }

  // Implements new #title_display value "placeholder"
  if (!empty($vars['element']['#title_display']) && $vars['element']['#title_display'] == 'placeholder') {
    $vars['attributes']['placeholder'] = $vars['element']['#title'];
  }
}

/**
 * Implements hook_field_widget_WIDGET_TYPE_form_alter(): datetime_default.
 */
function improvements_field_widget_datetime_default_form_alter(array &$element, FormStateInterface $form_state, array $context): void {
  // Remove wrapper from datetime widget
  if (isset($element['#theme_wrappers'])) {
    $element['#type'] = 'item';
    $element['#theme_wrappers'] = array_diff($element['#theme_wrappers'], ['fieldset']);
    $element['#theme_wrappers'][] = 'form_element';
  }
}

/**
 * Preprocess function for select.html.twig.
 */
function improvements_preprocess_select(array &$vars): void {
  // Create attributes for each option
  $options_attributes = $vars['element']['#options_attributes'] ?? [];

  foreach ($vars['options'] as &$option) {
    $option_attributes = [];

    if (isset($option['value'])) {
      $option_attributes = $options_attributes[$option['value']] ?? [];
      $option_attributes['value'] = $option['value'];
    }
    if (!empty($option['selected'])) {
      $option_attributes['selected'] = 'selected';
    }

    $option['attributes'] = new Attribute($option_attributes);

    // Fallback for themes with overriden select.html.twig
    if (isset($option['value'], $options_attributes[$option['value']])) {
      $option_attributes = new Attribute($options_attributes[$option['value']]);
      $option['value'] = Markup::create(Html::escape($option['value']) . '" ' . $option_attributes . ' data-drupal-empty="');
    }
  }
}
