<?php

use Drupal\Component\Utility\NestedArray;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Render\BubbleableMetadata;

/**
 * Implements hook_config_schema_info_alter().
 */
function improvements_config_schema_info_alter(array &$definitions): void {
  $definitions['system.site']['mapping']['improvements'] = [
    'type' => 'sequence',
    'label' => 'Third-party site information',
    'sequence' => [
      'type' => 'site_third_party_setting.[%key]',
    ],
  ];
}

/**
 * Implements hook_site_third_party_settings_info().
 */
function improvements_site_third_party_settings_info(): array {
  return [
    'company_name' => [
      'title' => t('Company name'),
      'form_element_type' => 'textfield',
      'weight' => 0,
    ],
    'mail_for_notifications' => [
      'title' => t('E-mail address for notifications'),
      'description' => t('E-mail address for notifications of new comments, contact messages and other.'),
      'form_element_type' => 'textfield',
      'weight' => 1,
    ],
    'phone' => [
      'title' => t('Phone number'),
      'form_element_type' => 'textfield',
      'weight' => 2,
    ],
  ];
}

/**
 * Return site third-party settings info.
 */
function improvements_get_site_third_party_settings_info(): ?array {
  static $cache;

  if ($cache === NULL) {
    $cache = \Drupal::moduleHandler()->invokeAll('site_third_party_settings_info');
  }

  return $cache;
}

/**
 * Implements hook_form_FORM_ID_alter(): system_site_information_settings.
 *
 * @see \Drupal\system\Form\SiteInformationForm
 */
function improvements_form_system_site_information_settings_alter(array &$form, FormStateInterface $form_state): void {
  $site_third_party_settings_info = improvements_get_site_third_party_settings_info();
  $site_information = improvements_get_site_information();

  foreach ($site_third_party_settings_info as $setting_name => $setting_info) {
    $group = $setting_info['group'] ?? 'site_information';
    $form_element_type = $setting_info['form_element_type'] ?? 'textfield';

    if (!isset($form[$group])) {
      $form[$group] = [
        '#type' => 'details',
        '#open' => TRUE,
      ];

      if (isset($setting_info['group_info'])) {
        $form[$group]['#title'] = $setting_info['group_info']['title'];
      }
    }

    $form[$group][$setting_name] = [
      '#type' => $form_element_type,
      '#title' => $setting_info['title'],
      '#description' => $setting_info['description'] ?? '',
      '#default_value' => $site_information[$setting_name] ?? NULL,
      '#weight' => $setting_info['weight'] ?? 0,
    ];
  }

  $form['#submit'][] = 'improvements_form_system_site_information_settings_submit';
}

/**
 * Custom submit callback for system_site_information_settings form.
 *
 * @see improvements_form_system_site_information_settings_alter()
 */
function improvements_form_system_site_information_settings_submit(array &$form, FormStateInterface $form_state): void {
  $site_information_editable_config = \Drupal::configFactory()->getEditable('system.site');
  $site_third_party_settings_info = improvements_get_site_third_party_settings_info();

  foreach ($site_third_party_settings_info as $setting_name => $setting_info) {
    $site_information_editable_config->set('improvements.' . $setting_name, $form_state->getValue($setting_name));
  }

  $site_information_editable_config->save();
}

/**
 * Return site information.
 *
 * @param string|array|null
 *
 * @return array|string
 */
function improvements_get_site_information($setting_key = NULL) {
  static $site_information;

  if ($site_information === NULL) {
    $site_information = \Drupal::config('system.site')->get();
    if (isset($site_information['improvements'])) {
      $site_information += $site_information['improvements'];
      unset($site_information['improvements']);
    }
  }

  if ($setting_key) {
    if (is_array($setting_key)) {
      return NestedArray::getValue($site_information, $setting_key);
    }
    return $site_information[$setting_key] ?? '';
  }

  return $site_information;
}

/**
 * Implements hook_filter_functions().
 */
function improvements_filter_functions(): array {
  return [
    'site_information' => [
      'function' => 'function_filter__site_information',
    ],
  ];
}

/**
 * Filter function returning site information.
 */
function function_filter__site_information(string $setting_key): string {
  if (str_contains($setting_key, '.')) {
    return improvements_get_site_information(explode('.', $setting_key));
  }
  return improvements_get_site_information($setting_key);
}

/**
 * @see improvements_token_info()
 */
function improvements_site_information_token_info(): array {
  $info = [];

  $site_third_party_settings_info = improvements_get_site_third_party_settings_info();
  foreach ($site_third_party_settings_info as $setting_name => $setting_info) {
    $info['tokens']['site'][$setting_name] = [
      'name' => $setting_info['title'],
      'description' => $setting_info['description'] ?? '',
    ];
  }

  return $info;
}

/**
 * @see improvements_tokens()
 */
function improvements_site_information_tokens(string $type, array $tokens, array $data, array $options, BubbleableMetadata $bubbleable_metadata): array {
  $replacements = [];

  // [site:*]
  if ($type == 'site') {
    $site_third_party_settings_info = improvements_get_site_third_party_settings_info();
    foreach ($tokens as $name => $original) {
      if (isset($site_third_party_settings_info[$name])) {
        $replacements[$original] = improvements_get_site_information($name);
      }
    }
  }

  return $replacements;
}
