<?php

namespace Drupal\improvements;

use Drupal\Component\Utility\Html;
use Drupal\Core\Render\Markup;
use Drupal\Core\Security\TrustedCallbackInterface;
use Drupal\Core\Url;

class ImprovementsTrustedCallbacks implements TrustedCallbackInterface {

  /**
   * {@inheritDoc}
   */
  public static function trustedCallbacks(): array {
    return [
      'pagePreRender',
      'linkPostRender',
      'managedFilePreRender',
      'contactMessageFormPreRender',
      'menuLinkContentFormPreRender',
    ];
  }

  /**
   * "Page" element pre render callback.
   *
   * @see improvements_element_info_alter()
   */
  public static function pagePreRender(array $element): array {
    // Remove "off canvas" wrapper for non-admin users.
    if (!\Drupal::currentUser()->hasPermission('administer site configuration')) {
      unset($element['#theme_wrappers']['off_canvas_page_wrapper']);
    }

    return $element;
  }

  /**
   * "Link" element post render callback.
   *
   * @see improvements_element_info_alter()
   */
  public static function linkPostRender($markup, array $element) {
    // Add support empty link fragment
    if (
      isset($element['#url']) &&
      $element['#url'] instanceof Url &&
      $element['#url']->getOption('fragment') == '<none>'
    ) {
      $markup = Markup::create(str_replace('href="#' . Html::escape('<none>') . '"', 'href="#"', $markup));
    }

    return $markup;
  }

  /**
   * "Managed file" element pre render callback.
   *
   * @see improvements_element_info_alter()
   */
  public static function managedFilePreRender(array $element): array {
    // Ability disable file auto-upload
    if (isset($element['upload']) && $element['upload']['#type'] == 'file') {
      if (isset($element['#autoupload']) && !$element['#autoupload']) {
        $element['upload']['#attributes']['class'][] = 'js-form-file--no-autoupload';
      }
    }

    // "Upload" button state
    // 1 - visible, 0 - removed, any other value - hidden
    if (isset($element['upload_button'], $element['#upload_button_state'])) {
      if ($element['#upload_button_state'] == 1) {
        $element['upload_button']['#attributes']['class'] = array_diff($element['upload_button']['#attributes']['class'], ['js-hide']);
      }
      elseif ($element['#upload_button_state'] == 0) {
        $element['upload_button']['#access'] = FALSE;
      }
    }

    return $element;
  }

  /**
   * Pre render callback for menu_link_content_form form.
   *
   * @see improvements_form_menu_link_content_form_alter()
   */
  public static function menuLinkContentFormPreRender(array $form): array {
    $form['description']['#weight'] = -1.1;
    return $form;
  }

}
